'''
This is an extension of lizard, that counts the complexity outside functions
'''


class LizardExtension(object):  # pylint: disable=R0903

    def __call__(self, tokens, reader):
        for token in tokens:
            yield token
        # adding the global pseudo function to the result
        reader.context.end_of_function()



import unittest
from lizard import analyze_file, FileAnalyzer, get_extensions
from languages import RubyReader


def get_ruby_function_list(source_code):
    return analyze_file.analyze_source_code(
        "a.rb", source_code).function_list

class Test_tokenizing_Ruby(unittest.TestCase):

    def check_tokens(self, expect, source):
        tokens = list(RubyReader.generate_tokens(source))
        self.assertEqual(expect, tokens)

    def test_tokenizing_ruby_regular_expression(self):
        self.check_tokens(['/ab/'], '/ab/')
        self.check_tokens([r'/\//'], r'/\//')
        self.check_tokens([r'/a/igm'], r'/a/igm')

    def test_should_not_confuse_division_as_regx(self):
        self.check_tokens(['a','/','b',',','a','/','b'], 'a/b,a/b')
        self.check_tokens(['3453',' ','/','b',',','a','/','b'], '3453 /b,a/b')

    def test_tokenizing_ruby_regular_expression(self):
        self.check_tokens(['a', '=', '/ab/'], 'a=/ab/')

    def test_tokenizing_ruby_comments(self):
        self.check_tokens(['/**a/*/'], '''/**a/*/''')

    def test_tokenizing_pattern(self):
        self.check_tokens(['/\//'], r'''/\//''')

    def test_tokenizing_string_expression(self):
        self.check_tokens(['%{"}'], r'''%{"}''')
        self.check_tokens(['%{""}'], r'''%{""}''')
        self.check_tokens(['%{\}}'], r'''%{\}}''')
        self.check_tokens(['%{\}}'], r'''%{\}}''')
        self.check_tokens(['%q{\}}'], r'''%q{\}}''')
        self.check_tokens(['%q[\]]'], r'''%q[\]]''')
        self.check_tokens(['%q<\>>'], r'''%q<\>>''')

    def test_vars(self):
        self.check_tokens(['$a'], r'''$a''')
        self.check_tokens(['@a'], r'''@a''')
        self.check_tokens(['@@a'], r'''@@a''')

    def test_ranges(self):
        self.check_tokens(['..'], r'''..''')
        self.check_tokens(['...'], r'''...''')

    def test_special_method_names(self):
        self.check_tokens(['a!'], r'''a!''')
        self.check_tokens(['a?'], r'''a?''')



class Test_parser_for_Ruby(unittest.TestCase):

    def test_empty(self):
        functions = get_ruby_function_list("")
        self.assertEqual(0, len(functions))

    def test_no_function(self):
        result = get_ruby_function_list(''' p "1" ''')
        self.assertEqual(0, len(result))

    def test_one_function(self):
        result = get_ruby_function_list('''
            def f
            end
                ''')
        self.assertEqual(1, len(result))
        self.assertEqual("f", result[0].name)
        self.assertEqual(0, result[0].parameter_count)
        self.assertEqual(1, result[0].cyclomatic_complexity)
        self.assertEqual(1, result[0].length)

    def test_one_function_loc(self):
        result = get_ruby_function_list('''
            def f
                something
            end
                ''')
        self.assertEqual(2, result[0].length)
        self.assertEqual(3, result[0].nloc)

    def test_two_functions(self):
        result = get_ruby_function_list('''
            def f
            end
            def g
            end
                ''')
        self.assertEqual(2, len(result))
        self.assertEqual("g", result[1].name)

    def test_one_with_begin_and_end(self):
        result = get_ruby_function_list('''
            def f
                begin
                    something
                end
            end
                ''')
        self.assertEqual(5, result[0].nloc)

    def test_one_with_begin_and_end_outside(self):
        result = get_ruby_function_list('''
        begin
            def f
                begin
                end
            end
        end
                ''')
        self.assertEqual(4, result[0].nloc)

    def test_one_with_class_in_it(self):
        result = get_ruby_function_list('''
            def f
                class a
                end
                module a
                end
            end
                ''')
        self.assertEqual(6, result[0].nloc)

    def test_one_with_class_as_identifier(self):
        result = get_ruby_function_list('''
            def f
                a.class
            end
                ''')
        self.assertEqual(3, result[0].nloc)

    def test_one_with_do(self):
        result = get_ruby_function_list('''
            def f
                x do
                    something
                end
            end
                ''')
        self.assertEqual(5, result[0].nloc)

    def test_one_within_do(self):
        result = get_ruby_function_list('''
            x do
                def f
                    something
                end
            end
                ''')
        self.assertEqual(3, result[0].nloc)

    def test_one_within_embedded_doc(self):
        result = get_ruby_function_list('''
=begin
    def f
    end
=end
                ''')
        self.assertEqual(0, len(result))

    def test_one_within_embedded_doc_harder(self):
        result = get_ruby_function_list('''
=begin
the everything between a line beginning with =begin and
that with =end will be skipped by the interpreter.
    def f
    end
=end
def f
end
                ''')
        self.assertEqual(1, len(result))


class Test_parser_for_Ruby_ccn(unittest.TestCase):

    def test_basic_complexity(self):
        result = get_ruby_function_list('''
            def f
                if a
                elsif b
                end
            end
                ''')
        self.assertEqual(3, result[0].cyclomatic_complexity)


class Test_parser_for_Ruby_if_while_for(unittest.TestCase):
    def test_basic_if_block(self):
        result = get_ruby_function_list('''
            def f
                if a
                end
            end
                ''')
        self.assertEqual(4, result[0].nloc)

    def test_basic_if_oneliner_block(self):
        result = get_ruby_function_list('''
            def f; if a; end; end
                ''')
        self.assertEqual(1, result[0].nloc)

    def test_basic_if_modifier(self):
        result = get_ruby_function_list('''
            def f
                a if b
            end
                ''')
        self.assertEqual(3, result[0].nloc)

    def test_basic_if_with_then_on_one_line(self):
        result = get_ruby_function_list('''
            def f
                a = if b then c else d end
            end
                ''')
        self.assertEqual(3, result[0].nloc)

    def test_unless(self):
        result = get_ruby_function_list('''
            def f
                a = unless b then c else d end
            end
                ''')
        self.assertEqual(3, result[0].nloc)

    def test_basic_while_block(self):
        result = get_ruby_function_list('''
            def f
                while a
                end
                for a
                end
            end
                ''')
        self.assertEqual(6, result[0].nloc)

    def test_basic_while_modifier(self):
        result = get_ruby_function_list('''
            def f
                a while a
            end
                ''')
        self.assertEqual(3, result[0].nloc)

    def test_while_with_do(self):
        result = get_ruby_function_list('''
            def f
                while a do
                end
            end
                ''')
        self.assertEqual(4, result[0].nloc)

    def test_while_modifier_with_block(self):
        result = get_ruby_function_list('''
            def f
                begin
                end while a
            end
                ''')
        self.assertEqual(4, result[0].nloc)


class Test_parser_for_Ruby_def(unittest.TestCase):
    def test_class_method(self):
        result = get_ruby_function_list('''
            def a.b
            end
                ''')
        self.assertEqual("a.b", result[0].name)
        self.assertEqual(0, result[0].parameter_count)

    def test_empty_parameters(self):
        result = get_ruby_function_list('''
            def a()
            end
                ''')
        self.assertEqual(0, result[0].parameter_count)

    def test_more_parameters(self):
        result = get_ruby_function_list('''
            def a(b,c)
            end
                ''')
        self.assertEqual(2, result[0].parameter_count)




'''
Language parser for JavaScript
'''

from lizard import CodeReader, CCppCommentsMixin
from .js_style_regex_expression import js_style_regex_expression


class JavaScriptReader(CodeReader, CCppCommentsMixin):
    # pylint: disable=R0903

    ext = ['js']
    language_names = ['javascript', 'js']

    @staticmethod
    @js_style_regex_expression
    def generate_tokens(source_code, _=''):
        return CodeReader.generate_tokens(source_code, _)

    def __init__(self, context):
        super(JavaScriptReader, self).__init__(context)
        # start from one, so global level will never count
        self.brace_count = 1
        self.last_tokens = ''
        self.function_name = ''
        self.function_stack = []

    def _state_global(self, token):
        if token == 'function':
            self._state = self._function
        elif token in ('=', ':'):
            self.function_name = self.last_tokens
        elif token in '.':
            self._state = self._field
            self.last_tokens += token
        else:
            if token == '{':
                self.brace_count += 1
            elif token == '}':
                self.brace_count -= 1
                if self.brace_count == 0:
                    self._state = self._state_global
                    self._pop_function_from_stack()
            self.last_tokens = token
            self.function_name = ''

    def _pop_function_from_stack(self):
        self.context.end_of_function()
        if self.function_stack:
            self.context.current_function = self.function_stack.pop()
            self.brace_count = self.context.current_function.brace_count

    def _function(self, token):
        if token != '(':
            self.function_name = token
        else:
            self.context.current_function.brace_count = self.brace_count
            self.function_stack.append(self.context.current_function)
            self.brace_count = 0
            self.context.start_new_function(self.function_name or 'function')
            self._state = self._dec

    def _field(self, token):
        self.last_tokens += token
        self._state = self._state_global

    def _dec(self, token):
        if token == ')':
            self._state = self._state_global
        else:
            self.context.parameter(token)
            return
        self.context.add_to_long_function_name(" " + token)



import unittest
from test.mock import patch
import lizard
from lizard import lizard_main
import os
import sys


@patch('lizard.md5_hash_file')
@patch('lizard.open', create=True)
@patch.object(os, 'walk')
@patch.object(lizard, 'print_result')
class TestApplication(unittest.TestCase):

    def testEmptyResult(self, print_result, os_walk, mock_open, _):

        def check_empty_result(result, options, scheme):
            self.assertEqual([], list(result))
            return 0

        os_walk.return_value = [('.', [], [])]
        print_result.side_effect = check_empty_result
        lizard_main(['lizard'])

    def testFilesWithFunction(self, print_result, os_walk, mock_open, _):
        def check_result(result, options, scheme):
            fileInfos = list(result)
            self.assertEqual(1, len(fileInfos))
            self.assertEqual('foo', fileInfos[0].function_list[0].name)
            return 0
        os_walk.return_value = [('.', [], ['a.cpp'])]
        mock_open.return_value.read.return_value = "void foo(){}"
        print_result.side_effect = check_result
        lizard_main(['lizard'])


class IntegrationTests(unittest.TestCase):

    def setUp(self):
        self.source_code = '''
        void foo() {
        #if
        #endif
            if(bar)
            {
            }
            switch(bar)
            {
              case 0: break;
              case 1: break;
              case 2: break;
              case 3: break;
            }
        }
        '''
        self.returned_warning_count = 0

    @patch('lizard.open', create=True)
    @patch.object(lizard, 'print_result')
    def run_with_mocks(self, argv, src, print_result, mock_open):
        def store_result(result, options, scheme):
            self.fileInfos = list(result)
            return self.returned_warning_count
        mock_open.return_value.read.return_value = src
        print_result.side_effect = store_result
        lizard_main(argv)
        return self.fileInfos

    @patch('lizard.md5_hash_file')
    @patch.object(os, 'walk')
    def runApplicationWithArgv(self, argv, os_walk, _):
        os_walk.return_value = [('.', [], ['a.cpp'])]
        return self.run_with_mocks(argv, self.source_code)

    def test_with_preprocessor_counted_in_CCN(self):
        self.runApplicationWithArgv(['lizard'])
        self.assertEqual(7, self.fileInfos[0].function_list[0].cyclomatic_complexity)

    def test_using_the_WordCount_plugin(self):
        self.runApplicationWithArgv(['lizard', '-EWordCount'])
        self.assertEqual(1, self.fileInfos[0].wordCount["foo"])

    def test_using_modified_ccn(self):
        self.runApplicationWithArgv(['lizard', '--modified'])
        self.assertEqual(4, self.fileInfos[0].function_list[0].cyclomatic_complexity)

    @patch.object(sys, 'exit')
    def test_exit_code(self, mock_exit):
        self.returned_warning_count = 6
        self.runApplicationWithArgv(['lizard', '-C5'])
        mock_exit.assert_called_with(1)




'''
This is an extension of lizard,
It helps to deal with C code with preprocessors that
is hard to parse. It works by always ignoring the code
between #else and #end.
'''


class LizardExtension(object):  # pylint: disable=R0903

    ordering_index = 0

    def __call__(self, tokens, reader):
        def preprocess_tokens(tokens):
            else_count = 0
            if_stack = []
            for token in tokens:
                if token.startswith("#"):
                    if token.startswith("#if"):
                        if_stack.append(token)
                    elif token.startswith("#el"):
                        else_count += 1
                        if_stack.append(token)
                    elif token.startswith("#endif"):
                        if if_stack and if_stack.pop().startswith("#el"):
                            else_count -= 1
                            if if_stack:
                                if_stack.pop()
                    continue
                if not else_count:
                    yield token

        if "c" not in reader.ext:
            return tokens
        else:
            return preprocess_tokens(tokens)



'''
This is an extension to lizard. It count the reccurance of every identifier
in the source code (ignoring the comments and strings), and then generate
a tag cloud based on the popularity of the identifiers.
The tag cloud is generated on an HTML5 canvas. So it will eventually save
the result to an HTML file and open the browser to show it.
'''

import webbrowser
from os.path import abspath
from lizard_ext.keywords import IGNORED_WORDS


class LizardExtension(object):

    HTML_FILENAME = "codecloud.html"
    ignoreList = IGNORED_WORDS

    def __init__(self):
        self.result = {}

    @staticmethod
    def __call__(tokens, reader):
        '''
        The function will be used in multiple threading tasks.
        So don't store any data with an extension object.
        '''
        reader.context.fileinfo.wordCount = result = {}
        for token in tokens:
            if token not in LizardExtension.ignoreList\
                    and token[0] not in ('"', "'", '#'):
                result[token] = result.get(token, 0) + 1
            yield token

    def reduce(self, fileinfo):
        '''
        Combine the statistics from each file.
        Because the statistics came from multiple thread tasks. This function
        needs to be called to collect the combined result.
        '''
        for k, val in fileinfo.wordCount.items():
            self.result[k] = self.result.get(k, 0) + val

    def print_result(self):
        with open(self.HTML_FILENAME, 'w') as html_file:
            html_file.write('''
<html>
    <head>
        <meta name="viewport" content="width=device-width,
            initial-scale=1.0,maximum-scale=1.0" />
        <style type="text/css">
            canvas {
                border: 1px solid black;
                width: 700px;
                height: 700px;
            }
        </style>
        <script type="text/javascript">
        ''')
            html_file.write(self.TAG_CLOUD_JAVASCRIPT)
            html_file.write('''
        </script>
        <script type="application/javascript">
            function draw() {
                var canvas = document.getElementById("canvas");
                    if (canvas.getContext) {
                        var ctx = canvas.getContext("2d");
                        // scale 2x
                        if(window.devicePixelRatio == 2) {
                            canvas.setAttribute('width', canvas.width * 2);
                            canvas.setAttribute('height', canvas.height * 2);
                        }
                        var tagCloud = new TagCloud(canvas.width,
                            canvas.height, ctx);
                        tagCloud.render([''')
            tags = sorted(self.result, key=self.result.get, reverse=True)[:400]
            for k in tags:
                html_file.write(
                    ' ' * 40 + '["%s", %d],\n' % (
                        k.replace('"', '\\\"')
                        .replace("'", "\\\\'").replace("\\", "\\\\"),
                        self.result[k]))
            html_file.write('''
                                    ]);
                                }
                        }
                    </script>
                </head>
                <body onload="draw();">
                    <canvas id="canvas" width="700" height="700"></canvas>
                </body>
            </html>''')

            webbrowser.open("file://" + abspath(self.HTML_FILENAME))

    TAG_CLOUD_JAVASCRIPT = '''

function TagCloud(w, h, context) {
    "use strict";
    this.ctx = context;
    this.canvasWidth = w;
    this.canvasHeight = h;
    this.fontSize = this.canvasHeight / 3;
    this.shape = "rectangle";
}

TagCloud.prototype.setShape = function () {
    this.shape = "circle";
};

TagCloud.prototype.render = function (tags) {
    this.ctx.textBaseline = "top";
    tags.forEach(function (tag) {
        this.placeTag(tag[0]);
    }, this);
};

TagCloud.prototype.placeTag = function (tag) {
    var placement;
    while (!(placement = this._getNonOverlappingPlaceWithBestSize(
            this.fontSize, tag)))
        this.fontSize *= 0.9;

    this.ctx.fillStyle = this._getRandomColor();
    this.ctx.fillText(tag, placement.x, placement.y);
};

TagCloud.prototype._getNonOverlappingPlaceWithBestSize =
    function (fontSize, tag) {
    this.ctx.font = "" + fontSize + "pt " + "Arial";
    var lineHeight=this.getLineHeight(fontSize);
    var tagWidth = this.ctx.measureText(tag).width;

    var base = new BasePlacement(
        (this.canvasWidth - tagWidth) * Math.random(),
        (this.canvasHeight - lineHeight) * Math.random(),
        lineHeight
        );

    var placement;
    /* jshint ignore:start */
    while (placement = base.nextPlaceToTry()) {
        if (this._isPlaceEmpty(placement, tagWidth, lineHeight))
            break;
    }
    /* jshint ignore:end */
    return placement;
};

TagCloud.prototype.getLineHeight = function (fontSize) {
    return this.ctx.measureText('M').width * 1.2;
}

TagCloud.prototype._getRandomColor = function (){
    var colors = ["aqua", "black", "blue", "fuchsia", "gray", "green",
                  "lime", "maroon", "navy", "olive", "orange", "purple",
                  "red", "silver", "teal"];
    return colors[Math.floor(colors.length * Math.random())];
};

TagCloud.prototype._isPlaceEmpty = function (placement, width, height) {
    if (placement.x < 0 || placement.y < 0 || placement.x + width >
         this.canvasWidth || placement.y + height > this.canvasHeight)
        return false;

    var pix = this.ctx.getImageData(
                placement.x, placement.y, width, height).data;

    for (var i = 0, n = pix.length; i < n; i += 4)
        if (pix[i+3])
                return false;

    return [[placement.x, placement.y],
            [placement.x + width, placement.y],
            [placement.x, placement.y + height],
            [placement.x + width, placement.y + height]].every(
                function(pos) {
                    var a = this.canvasWidth / 2;
                    var b = this.canvasHeight / 2;
                    var X = pos[0] - a;
                    var Y = pos[1] - b;
                    return (X * X / a / a + Y * Y / b / b < 1);
                }, this);
};

TagCloud.prototype.getCoverage = function () {
    var pix = this.ctx.getImageData(
                0, 0, this.canvasWidth, this.canvasHeight).data;
    var pixCount = 0;
    for (var i = 0, n = pix.length; i < n; i += 4) {
        if (pix[i+3])
            pixCount++;
    }
    return pixCount * 100 / this.canvasWidth / this.canvasHeight;
};

function BasePlacement(x, y, h) {
    var baseX = x,
        baseY = y,
        scale = h,
        tryNumber = 0;

    this.nextPlaceToTry = function() {
        if (tryNumber < this._spiralOffsets.length)
            return {
                x : baseX + this._spiralOffsets[tryNumber][0] * scale,
                y : baseY + this._spiralOffsets[tryNumber++][1] * scale
            };
    };
}

function generateSpiralOffsets() {
    var spiralOffsets = [];
    var radius = 0;
    var dr = 0.2;
    for (var i = 0; radius < 40; i+=0.4, radius += dr) {
        spiralOffsets.push([
                   radius * Math.sin(i),
                   radius * Math.cos(i)
                ]);
    }
    return spiralOffsets;
}

BasePlacement.prototype._spiralOffsets = generateSpiralOffsets();

    '''



import unittest
from lizard import parse_args
from test.mock import patch
import sys
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class TestOptionParsing(unittest.TestCase):

    def test_should_use_current_folder_as_default_path(self):
        options = parse_args(['lizard'])
        self.assertEqual(['.'], options.paths)

    def test_default_sorting(self):
        options = parse_args(['lizard'])
        self.assertEqual(0, len(options.sorting))

    def test_sorting_factor(self):
        options = parse_args(['lizard', '-snloc'])
        self.assertEqual("nloc", options.sorting[0])

    def test_threshold(self):
        options = parse_args(['lizard', '-Tnloc=123'])
        self.assertEqual(123, options.thresholds['nloc'])

    def test_more_threshold(self):
        options = parse_args(['lizard', '-Tnloc=123', '-T length = 300'])
        self.assertEqual(300, options.thresholds['length'])

    def test_ccn(self):
        options = parse_args(['lizard', '-C123'])
        self.assertEqual(123, options.thresholds['cyclomatic_complexity'])

    def test_length(self):
        options = parse_args(['lizard', '-L123'])
        self.assertEqual(123, options.thresholds['length'])

    def test_arguments(self):
        options = parse_args(['lizard', '-a123'])
        self.assertEqual(123, options.thresholds['parameter_count'])

    @patch.object(sys, 'exit')
    @patch('sys.stderr')
    def test_sorting_factor_does_not_exist(self, _, mock_exit):
        options = parse_args(['lizard', '-sdoesnotexist'])
        mock_exit.assert_called_with(2)

    @patch.object(sys, 'exit')
    @patch('sys.stderr', new_callable=StringIO)
    def test_sorting_factor_does_not_exist_error_message(self, mock_stderr, mock_exit):
        options = parse_args(['lizard', '-sdoesnotexist'])
        self.assertEqual("Wrong field name 'doesnotexist'.\nCandidates are: nloc, cyclomatic_complexity, token_count, parameter_count, length, location\n", mock_stderr.getvalue())

    @patch.object(sys, 'exit')
    @patch('sys.stderr')
    def test_sorting_factor_does_not_exist(self, _, mock_exit):
        options = parse_args(['lizard', '-Tdoesnotexist=3'])
        mock_exit.assert_called_with(2)




''' extensions of lizard '''

from __future__ import print_function
from .htmloutput import html_output
from .xmloutput import xml_output


def print_xml(results, options, _):
    print(xml_output(list(results), options.verbose))
    return 0



import unittest
from lizard import CodeReader, CLikeReader
from languages import JavaReader, ObjCReader, JavaScriptReader


class TestLanguageChooser(unittest.TestCase):

    def test_not_case_sensitive(self):
        self.assertEqual(CLikeReader, CodeReader.get_reader("a.Cpp"))

    def test_java(self):
        self.assertEqual(JavaReader, CodeReader.get_reader("a.java"))

    def test_objectiveC(self):
        self.assertEqual(ObjCReader, CodeReader.get_reader("a.m"))

    def test_c_cpp(self):
        for name in ("a.cpp", ".cxx", ".h", ".hpp"):
            self.assertEqual(CLikeReader, CodeReader.get_reader(name),
                             "File name '%s' is not recognized as c/c++ file" % name);

    def test_JavaScript(self):
        self.assertEqual(JavaScriptReader, CodeReader.get_reader("a.js"))

    def test_unknown_extension(self):
        self.assertEqual(None, CodeReader.get_reader("a.unknown"));




import unittest
from test.mock import Mock, patch
from .testHelpers import get_cpp_fileinfo, get_cpp_function_list

class Test_Token_Count(unittest.TestCase):

    def test_non_function_tokens_are_counted(self):
        fileinfo = get_cpp_fileinfo("int i, j;")
        self.assertEqual(5, fileinfo.token_count)

    def test_include_is_counted_as_2(self):
        fileinfo = get_cpp_fileinfo("#include \"abc.h\"")
        self.assertEqual(2, fileinfo.token_count)

    def test_include_with_lg_and_gg_is_counted_as_2(self):
        fileinfo = get_cpp_fileinfo("#include <abc.h>")
        self.assertEqual(2, fileinfo.token_count)

    def test_one_function_with_no_token(self):
        result = get_cpp_function_list("int fun(){}")
        self.assertEqual(5, result[0].token_count)

    def test_one_function_with_one_token(self):
        result = get_cpp_function_list("int fun(){;}")
        self.assertEqual(6, result[0].token_count)

    def test_one_function_with_content(self):
        result = get_cpp_function_list("int fun(){if(a){xx;}}")
        self.assertEqual(13, result[0].token_count)

    def test_one_function_with_comments_only(self):
        result = get_cpp_function_list("int fun(){/**/}")
        self.assertEqual(5, result[0].token_count)

class TestNLOC(unittest.TestCase):

    def test_one_function_with_content(self):
        result = get_cpp_function_list("int fun(){if(a){xx;}}")
        self.assertEqual(1, result[0].nloc)

    def test_nloc_of_empty_function(self):
        result = get_cpp_function_list("int fun(){}")
        self.assertEqual(1, result[0].nloc)

    def test_nloc(self):
        result = get_cpp_function_list("int fun(){\n\n\n}")
        self.assertEqual(2, result[0].nloc)

    def test_nloc_with_new_line_in_comment(self):
        result = get_cpp_function_list("int fun(){/*\n*/}")
        self.assertEqual(2, result[0].nloc)

    def test_nloc_with_comment_between_new_lines(self):
        result = get_cpp_function_list("int fun(){\n/*\n*/\n}")
        self.assertEqual(2, result[0].nloc)

    def test_nloc2(self):
        result = get_cpp_function_list("int fun(){aa();\n\n\n\nbb();\n\n\n}")
        self.assertEqual(3, result[0].nloc)
        self.assertEqual(1, result[0].start_line)
        self.assertEqual(8, result[0].end_line)

    def check_file_nloc(self, expect, source):
        fileinfo = get_cpp_fileinfo(source)
        self.assertEqual(expect, fileinfo.nloc)

    def test_last_line_without_return_should_be_counted_in_fileinfo(self):
        self.check_file_nloc(1, ";\n")
        self.check_file_nloc(2, ";\n\n;\n")
        self.check_file_nloc(2, ";\n;")
        self.check_file_nloc(1, "fun(){}")
        self.check_file_nloc(1, "fun(){};\n")


class TestLOC(unittest.TestCase):

    def test_having_empty_line(self):
        result = get_cpp_function_list("\nint fun(){}")
        self.assertEqual(2, result[0].start_line)

    def test_newline_in_macro(self):
        result = get_cpp_function_list("#define a\\\nb\nint fun(){}")
        self.assertEqual(3, result[0].start_line)

    def test_having_empty_line_that_has_spaces(self):
        result = get_cpp_function_list("  \nint fun(){}")
        self.assertEqual(2, result[0].start_line)

    def test_having_multiple_line_comments(self):
        result = get_cpp_function_list('''int fun(){
        /*2
          3
          4*/
                }''')
        self.assertEqual(5, result[0].end_line)

class TestFileNLOC(unittest.TestCase):

    def test_empty_file_should_has_0_nloc(self):
        fileinfo = get_cpp_fileinfo("")
        self.assertEqual(0, fileinfo.nloc)

    def test_one_line_file_should_has_1_nloc(self):
        fileinfo = get_cpp_fileinfo("a")
        self.assertEqual(1, fileinfo.nloc)

    def test_one_line_file_with_newline_at_the_end_should_has_1_nloc(self):
        fileinfo = get_cpp_fileinfo("a\n")
        self.assertEqual(1, fileinfo.nloc)

    def test_two_one_line_file_should_has_2_nloc(self):
        fileinfo = get_cpp_fileinfo("a\nb")
        self.assertEqual(2, fileinfo.nloc)

    def test_comment_should_not_count(self):
        fileinfo = get_cpp_fileinfo("//abc")
        self.assertEqual(0, fileinfo.nloc)



import unittest
import sys
from test.mock import Mock, patch
from test.helper_stream import StreamStdoutTestCase
import os
from lizard import print_warnings, print_and_save_modules, FunctionInfo, FileInformation,\
    print_result, get_extensions, OutputScheme, get_warnings, print_clang_style_warning,\
    parse_args
from lizard_ext import xml_output

def print_result_with_scheme(result, option):
    return print_result(result, option, OutputScheme(option.extensions))

class TestFunctionOutput(StreamStdoutTestCase):

    def setUp(self):
        StreamStdoutTestCase.setUp(self)
        self.extensions = get_extensions([])
        self.scheme = OutputScheme(self.extensions)
        self.foo = FunctionInfo("foo", 'FILENAME', 100)

    def test_function_info_header_should_have_a_box(self):
        print_and_save_modules([], self.extensions, self.scheme)
        self.assertIn("=" * 20, sys.stdout.stream.splitlines()[0])

    def test_function_info_header_should_have_the_captions(self):
        print_and_save_modules([], self.extensions, self.scheme)
        self.assertEquals("  NLOC    CCN   token  PARAM  length  location  ", sys.stdout.stream.splitlines()[1])

    def test_function_info_header_should_have_the_captions_of_external_extensions(self):
        external_extension = Mock(FUNCTION_CAPTION = "*external_extension*", ordering_index=-1)
        extensions = get_extensions([external_extension])
        scheme = OutputScheme(extensions)
        print_and_save_modules([], extensions, scheme)
        self.assertEquals("  NLOC    CCN   token  PARAM  length *external_extension* location  ", sys.stdout.stream.splitlines()[1])

    def test_print_fileinfo(self):
        self.foo.end_line = 100
        self.foo.cyclomatic_complexity = 16
        fileStat = FileInformation("FILENAME", 1, [self.foo])
        print_and_save_modules([fileStat], self.extensions, self.scheme)
        self.assertEquals("       1     16      1      0       0 foo@100-100@FILENAME", sys.stdout.stream.splitlines()[3])

class TestWarningOutput(StreamStdoutTestCase):

    def setUp(self):
        StreamStdoutTestCase.setUp(self)
        self.option = parse_args("app")
        self.foo = FunctionInfo("foo", 'FILENAME', 100)
        fileSummary = FileInformation("FILENAME", 123, [self.foo])
        self.scheme = Mock()

    def test_should_have_header_when_warning_only_is_off(self):
        print_warnings(self.option, self.scheme, [])
        self.assertIn("cyclomatic_complexity > 15", sys.stdout.stream)

    def test_no_news_is_good_news(self):
        count = print_clang_style_warning([], self.option, None)
        self.assertEqual('', sys.stdout.stream)
        self.assertEqual(0, count)

    def test_should_use_clang_format_for_warning(self):
        self.foo.cyclomatic_complexity = 30
        fileSummary = FileInformation("FILENAME", 123, [self.foo])
        count = print_clang_style_warning([fileSummary], self.option, None)
        self.assertIn("FILENAME:100: warning: foo has 30 CCN and 0 params (1 NLOC, 1 tokens)\n", sys.stdout.stream)
        self.assertEqual(1, count)

    def test_sort_warning(self):
        self.option.sorting = ['cyclomatic_complexity']
        self.foo.cyclomatic_complexity = 30
        bar = FunctionInfo("bar", '', 100)
        bar.cyclomatic_complexity = 40
        fileSummary = FileInformation("FILENAME", 123, [self.foo, bar])
        warnings = get_warnings([fileSummary], self.option)
        self.assertEqual('bar', warnings[0].name)

    def test_sort_warning_with_generator(self):
        self.option.sorting = ['cyclomatic_complexity']
        print_warnings(self.option, self.scheme, (x for x in []))


class TestFileOutput(StreamStdoutTestCase):

    def test_print_and_save_detail_information(self):
        fileSummary = FileInformation("FILENAME", 123, [])
        print_and_save_modules([fileSummary], [], Mock())
        self.assertIn("    123      0    0.0         0         0     FILENAME", sys.stdout.stream)

    def test_print_file_summary_only_once(self):
        print_and_save_modules(
                            [FileInformation("FILENAME1", 123, []),
                             FileInformation("FILENAME2", 123, [])], [], Mock())
        self.assertEqual(1, sys.stdout.stream.count("FILENAME1"))


class TestAllOutput(StreamStdoutTestCase):

    def setUp(self):
        StreamStdoutTestCase.setUp(self)
        self.foo = FunctionInfo("foo", 'FILENAME', 100)

    def test_print_extension_results(self):
        file_infos = []
        extension = Mock(FUNCTION_CAPTION = "")
        option = Mock(CCN=15, thresholds={}, number = 0, extensions = [extension], whitelist='')
        print_result(file_infos, option, OutputScheme(option.extensions))
        self.assertEqual(1, extension.print_result.call_count)

    def test_should_not_print_extension_results_when_not_implemented(self):
        file_infos = []
        option = Mock(CCN=15, number = 0, thresholds={}, extensions = [object()], whitelist='')
        return print_result_with_scheme(file_infos, option)

    def test_print_result(self):
        file_infos = [FileInformation('f1.c', 1, []), FileInformation('f2.c', 1, [])]
        option = Mock(CCN=15,thresholds={},  number = 0, extensions=[], whitelist='')
        self.assertEqual(0, print_result_with_scheme(file_infos, option))

    @patch.object(os.path, 'isfile')
    @patch('lizard.open', create=True)
    def check_whitelist(self, script, mock_open, mock_isfile):
        mock_isfile.return_value = True
        mock_open.return_value.read.return_value = script
        file_infos = [FileInformation('f1.c', 1, [self.foo])]
        option = Mock(thresholds={'cyclomatic_complexity':15, 'length':1000}, CCN=15, number = 0, arguments=100, length=1000, extensions=[])
        return print_result_with_scheme(file_infos, option)

    def test_exit_with_non_zero_when_more_warning_than_ignored_number(self):
        self.foo.cyclomatic_complexity = 16
        self.assertEqual(1, self.check_whitelist(''))

    def test_whitelist(self):
        self.foo.cyclomatic_complexity = 16
        self.check_whitelist('foo')
        self.assertEqual(0, self.check_whitelist('foo'))

    def test_null_result(self):
        self.check_whitelist('')


class TestXMLOutput(unittest.TestCase):
    foo = FunctionInfo("foo", '', 100)
    foo.cyclomatic_complexity = 16
    file_infos = [FileInformation('f1.c', 1, [foo])]
    xml = xml_output(file_infos, True)

    def test_xml_output(self):
        self.assertIn('''foo at f1.c:100''', self.xml)

    def test_xml_stylesheet(self):
        self.assertIn('''<?xml-stylesheet type="text/xsl" href="https://raw.github.com/terryyin/lizard/master/lizard.xsl"?>''', self.xml)



'''
Common behaviours of script languages
'''


class ScriptLanguageMixIn(object):
    # pylint: disable=R0903

    @staticmethod
    def get_comment_from_token(token):
        if token.startswith("#"):
            return token[1:]



import unittest
from .testHelpers import get_cpp_function_list_with_extnesion
from lizard_ext.lizardignoreassert import LizardExtension

class TestAssersionExtension(unittest.TestCase):
    def test_exclusion_of_assert(self):
        result = get_cpp_function_list_with_extnesion("void fun() { assert(a && b && c); }", LizardExtension())
        self.assertEqual(1, result[0].cyclomatic_complexity)

    def test_exclusion_of_static_assert(self):
        result = get_cpp_function_list_with_extnesion(
                "void fun() { static_assert(a && b && c, \"Failed\"); }", LizardExtension())
        self.assertEqual(1, result[0].cyclomatic_complexity)

    def test_assert_with_ternary_operator(self):
        result = get_cpp_function_list_with_extnesion(
                "void fun() { assert(a ? b : c); }",
                LizardExtension())
        self.assertEqual(1, result[0].cyclomatic_complexity)

    def test_ignore_all_code_in_assert(self):
        result = get_cpp_function_list_with_extnesion(
                """void fun() { assert(any_of(v.begin(), v.end(),
                                       [](auto& a) { return a ? b : c })); }""",
                LizardExtension())
        self.assertEqual(1, result[0].cyclomatic_complexity)



from test.mock import Mock, patch
import unittest
import sys
from lizard_ext import html_output
from lizard import parse_args, FunctionInfo, FileInformation
from test.helper_stream import StreamStdoutTestCase


class TestHTMLOutput(StreamStdoutTestCase):

    def setUp(self):
        StreamStdoutTestCase.setUp(self)
        self.option = parse_args("app")
        self.foo = FunctionInfo("foo", 'FILENAME', 100)
        self.fileSummary = FileInformation("FILENAME", 123, [self.foo])
        self.scheme = Mock()

    def test_should_have_html_body(self):
        html_output([self.fileSummary], self.option)
        self.assertRegexpMatches(sys.stdout.stream,
                r"<html>.*<body>.*</body>\s*</html>\s*")



import unittest
from .testHelpers import get_cpp_function_list_with_extnesion
from lizard_ext.lizardinternalrelationship import LizardExtension as ir
from lizard_ext.lizardinternalrelationship import preprocess


def TestInternalRelationship():#(unittest.TestCase):

    def test_empty_source(self):
        result = get_cpp_function_list_with_extnesion("", ir())
        self.assertEqual(0, len(result))

    def test_a_c_function(self):
        result = get_cpp_function_list_with_extnesion("void f(){}", ir())
        self.assertEqual(1, len(result))
        self.assertEqual(set(), result[0].dependencies)

    def test_a_function_with_one_dependency(self):
        result = get_cpp_function_list_with_extnesion("int C::f(){a();}", ir())
        self.assertEqual({"a"}, result[0].dependencies)

    def test_a_function_with_keyword(self):
        result = get_cpp_function_list_with_extnesion("int C::f(){if(m)b();}", ir())
        self.assertEqual({"m", "b"}, result[0].dependencies)

    def test_with_local_variable(self):
        result = get_cpp_function_list_with_extnesion("int C::f(){int a;}", ir())
        self.assertEqual(set(), result[0].dependencies)

    def test_using_local_variable(self):
        result = get_cpp_function_list_with_extnesion("int C::f(){type a; a;}", ir())
        self.assertEqual(set(), result[0].dependencies)

    def test_scoped_name(self):
        result = get_cpp_function_list_with_extnesion("int C::f(){a::b();}", ir())
        self.assertEqual({"a::b"}, result[0].dependencies)

    def test_parameter(self):
        result = get_cpp_function_list_with_extnesion("void f(type a){a;}", ir())
        self.assertEqual(set(), result[0].dependencies)

    def xtest_xxx(self):
        result = get_cpp_function_list_with_extnesion("""
void CmDiagnosticsService::startService()
{
    TRACE_LOG_BEG(log_);
    auto func = limFacade_->fetchByClass(lim::MoFunc_l::className);

    if (!func)
    {
        INFO_LOG(log_) << "Failed to retrieve Func_L object.";
        limFacade_->subscribeForFunclCreation(
            core::types::bindWeak(shared_from_this(), &CmDiagnosticsService::funcCreateHandler),
            subscriptionFuncl_);
        return;
    }
    createAndInitializeCmDiagnostics(func);
    TRACE_LOG_END(log_);
}""", ir())
        self.assertEqual({"a::b"}, result[0].dependencies)

class TestPreprocessor(unittest.TestCase):

    def process(self, *tokens):
        return [x for x in preprocess(tokens)]

    def test_no_namespaces(self):
        self.assertEqual(["a", "b"], self.process("a", "b"))

    def test_global_namespaces(self):
        self.assertEqual(["::b"], self.process("::", "b"))

    def test_namespaces(self):
        self.assertEqual(["a::b"], self.process("a", "::", "b"))

    def test_namespaces_and_no_namepace(self):
        self.assertEqual(["a::b", "c"], self.process("a", "::", "b", "c"))




'''
This is an extension of lizard, that ignores the CCN within
the assertion.
'''
from lizard import CodeStateMachine


class LizardExtension(CodeStateMachine):  # pylint: disable=R0903

    def __init__(self):
        super(LizardExtension, self).__init__(None)

    def _state_global(self, token):
        if token in ("assert", "static_assert"):
            self._state = self.in_assertion

    @CodeStateMachine.read_inside_brackets_then("()", "_state_global")
    def in_assertion(self, token):
        if token in ("&&", "||", "?"):
            self.context.add_condition(-1)



import unittest
from test.mock import Mock, patch
from lizard import get_extensions
import importlib

class FakeExtension:
    def LizardExtension(self):
        return "fake extension"

@patch.object(importlib, 'import_module')
class Test_mounting_extensions(unittest.TestCase):

    def test_should_append_extension_at_the_end_by_default(self, mock_import):
        mock_import.return_value = FakeExtension()
        exts = get_extensions(["my_ext"])
        self.assertEqual("fake extension", exts[-1])

    def test_should_insert_extension_at_the_index_when_specified(self, mock_import):
        extension = Mock(ordering_index=1)
        mock_import.return_value = extension
        exts = get_extensions([extension])
        self.assertEqual(extension, exts[1])




'''
Language parser for JavaScript
'''

from lizard import CodeReader
from .script_language import ScriptLanguageMixIn
from .js_style_regex_expression import js_style_regex_expression


class SyntaxMachine(object):
    # pylint: disable=R0903
    def __init__(self, context):
        self.context = context
        self.saved_state = self._state = self._state_global
        self.last_token = None
        self.to_exit = False
        self.callback = None

    def next(self, state, token=None):
        self._state = state
        if token is not None:
            return self(token)

    def sm_return(self):
        self.to_exit = True

    def sub_state(self, state, callback=None, token=None):
        self.saved_state = self._state
        self.callback = callback
        self.next(state, token)

    def __call__(self, token):
        if self._state(token):
            self.next(self.saved_state)
            if self.callback:
                self.callback()
        self.last_token = token
        if self.to_exit:
            return True

    def _state_global(self, token):
        pass


def state_embedded_doc(token):
    if token == "=end":
        return True


class RubyStateMachine(SyntaxMachine):
    # pylint: disable=R0903
    def is_newline(self):
        return self.context.newline or self.last_token == ";"

    def _state_global(self, token):
        if token == "end":
            self.sm_return()
        elif token == 'def':
            self._state = self._def
            self.next(self._def)
        elif token in ("begin", "do", "class", "module"
                       ) and self.last_token != ".":
            self.sub_state(RubyStateMachine(self.context))
        elif token in ("while", "for"):
            if self.is_newline():
                self.next(self._for_while)
        elif token in ("if", "unless"):
            if self.is_newline():
                self.sub_state(RubyStateMachine(self.context))
            else:
                self.next(self._if)
        elif token == "=begin":
            self.sub_state(state_embedded_doc)

    def _def(self, token):
        self.context.start_new_function(token)
        self.next(self._def_continue)

    def _def_continue(self, token):
        def callback():
            self.context.end_of_function()
            self.next(self._state_global)
        if token == ".":
            self.context.add_to_function_name(token)
            self.next(self._def_class_method)
        elif token == "(":
            self.next(self._def_parameters)
        else:
            self.sub_state(RubyStateMachine(self.context), callback, token)

    def _def_class_method(self, token):
        self.context.add_to_function_name(token)
        self.next(self._def_continue)

    def _def_parameters(self, token):
        if token == ')':
            self.next(self._def_continue)
        else:
            self.context.parameter(token)
            return
        self.context.add_to_long_function_name(" " + token)

    def _if(self, token):
        if self.is_newline():
            self.next(self._state_global, token)
        elif token == "then":
            self.next(self._state_global)
            self.sub_state(RubyStateMachine(self.context))

    def _for_while(self, token):
        if self.is_newline() or token == "do":
            self.next(self._state_global)
            if token != "end":
                self.sub_state(RubyStateMachine(self.context))


class RubyReader(CodeReader, ScriptLanguageMixIn):
    # pylint: disable=R0903

    ext = ['rb']
    language_names = ['ruby']
    conditions = set(['if', 'until', 'for', 'while', 'and', 'or',
                      'elsif', 'rescue', 'ensure', 'when', '||', '&&', '?'])

    def __init__(self, context):
        super(RubyReader, self).__init__(context)
        self._state = RubyStateMachine(context)

    @staticmethod
    @js_style_regex_expression
    def generate_tokens(source_code, _=''):
        return CodeReader.generate_tokens(
            source_code,
            r"|^\=begin|^\=end" +
            r"|\%[qQr]?\{(?:\\.|[^\}\\])*?\}" +
            r"|\%[qQr]?\[(?:\\.|[^\]\\])*?\]" +
            r"|\%[qQr]?\<(?:\\.|[^\>\\])*?\>" +
            r"|\$\w+" +
            r"|\@\w+" +
            r"|\@\@\w+" +
            r"|\.+" +
            r"|\w+\?" +
            r"|\w+\!" +
            _)



''' programming languages of lizard '''

from .script_language import ScriptLanguageMixIn
from .java import JavaReader
from .javascript import JavaScriptReader
from .python import PythonReader
from .objc import ObjCReader
from .ttcn import TTCNReader
from .swift import SwiftReader
from .ruby import RubyReader



'''
generate token with javascript style regular expression.
'''

import re


def js_style_regex_expression(func):
    def generate_tokens_with_regex(source_code, _=""):
        regx_regx = r"|/(?:\\/|[^/])+?/[igm]*"
        regx_pattern = re.compile(regx_regx)
        word_pattern = re.compile(r'\w+')
        tokens = func(source_code, regx_regx)
        leading_by_word = False
        for token in tokens:
            if leading_by_word and regx_pattern.match(token):
                for subtoken in func(token, _):
                    yield subtoken
            else:
                yield token
            if not token.isspace():
                leading_by_word = word_pattern.match(token)
    return generate_tokens_with_regex



from lizard import lizard_main
import languages
import sys

def lizard_for_more_languages(args):
    lans = [languages.JavaReader,
        languages.JavaScriptReader,
        languages.PythonReader,
        languages.ObjCReader,
        languages.TTCNReader,
        languages.RubyReader,
        languages.SwiftReader]
    lizard_main(args, lans)

if __name__ == "__main__":
    lizard_for_more_languages(sys.argv)



#
# Unit Test
#
import unittest
import sys
from mock import patch, Mock
from lizard import CLikeReader, map_files_to_analyzer, FunctionInfo, analyze_file, FileInfoBuilder


def analyzer_mock(filename):
    return filename

class Test_analyze_files(unittest.TestCase):
    def test_NoFiles(self):
        call_count = 0
        def analyzer(filename):
            call_count += 1
        map_files_to_analyzer([], analyzer, 1)
        self.assertEqual(0, call_count)

    def test_NoFilesMultipleThread(self):
        call_count = 0
        def analyzer(filename):
            call_count += 1
        map_files_to_analyzer([], analyzer, 2)
        self.assertEqual(0, call_count)
        
    def test_OneFile(self):
        analyzer = analyzer_mock
        r = map_files_to_analyzer(["filename"], analyzer, 1)
        self.assertEqual(["filename"], [x for x in r])
        
    def test_OneFileMultipleThread(self):
        analyzer = analyzer_mock
        r = map_files_to_analyzer(["filename"], analyzer, 2)
        self.assertEqual(["filename"], [x for x in r])
    
    def test_MoreFiles(self):
        analyzer = analyzer_mock
        r = map_files_to_analyzer(["f1", "f2"], analyzer, 1)
        self.assertEqual(["f1", "f2"], [x for x in r])

    def test_MoreFilesMultipleThread(self):
        analyzer = analyzer_mock
        r = map_files_to_analyzer(["f1", "f2"], analyzer, 2)
        self.assertSetEqual(set(["f1", "f2"]), set(x for x in r))


@patch('lizard.open', create=True)
class Test_FileAnalyzer(unittest.TestCase):
    
    def setUp(self):
        self.analyzer = analyze_file
        
    def test_analyze_c_file(self, mock_open):
        file_handle = mock_open.return_value.read.return_value = "int foo(){haha();\n}"
        r = map_files_to_analyzer(["f1.c"], self.analyzer, 1)
        self.assertEqual(1, len([x for x in r]))
        
    def test_analyze_c_file_with_multiple_thread(self, mock_open):
        file_handle = mock_open.return_value.read.return_value = "int foo(){haha();\n}"
        r = map_files_to_analyzer(["f1.c"], self.analyzer, 2)
        self.assertEqual(1, len([x for x in r]))
    
    def test_fileInfomation(self, mock_open):
        mock_open.return_value.read.return_value = "int foo(){haha();\n}"
        r = map_files_to_analyzer(["f1.c"], self.analyzer, 1)
        fileInfo = list(r)[0]
        self.assertEqual(2, fileInfo.nloc)
        self.assertEqual(2, fileInfo.average_NLOC)
        self.assertEqual(1, fileInfo.average_CCN)
        self.assertEqual(9, fileInfo.average_token)

    @patch.object(sys, 'stderr')
    def test_should_report_when_having_other_problem_and_continue(self, mock_stderr, mock_open):
        mock_open.side_effect = IOError("[Errno 2] No such file or directory")
        analyze_file("f1.c")
        self.assertEqual(1, mock_stderr.write.call_count)
        error_message = mock_stderr.write.call_args[0][0]
        self.assertEqual("Error: Fail to read source file 'f1.c'\n", error_message)

class Test_Picklability(unittest.TestCase):

    def test_FunctionInfo_ShouldBePicklable(self):
        import pickle
        pickle.dumps(FunctionInfo("a", '', 1))

    def test_FileInfo_ShouldBePicklable(self):
        import pickle
        pickle.dumps(FileInfoBuilder("a"))


from lizard import warning_filter, FileInformation, whitelist_filter

class TestWarningFilter(unittest.TestCase):

    def setUp(self):
        complex_fun = FunctionInfo("complex", '', 100)
        complex_fun.cyclomatic_complexity = 16
        simple_fun = FunctionInfo("simple", '', 100)
        simple_fun.cyclomatic_complexity = 15
        self.fileStat = FileInformation("FILENAME", 1, [complex_fun, simple_fun])

    def test_should_filter_the_warnings(self):
        option = Mock(thresholds={'cyclomatic_complexity':15, 'length':1000})
        warnings = list(warning_filter(option, [self.fileStat]))
        self.assertEqual(1, len(warnings))
        self.assertEqual("complex", warnings[0].name)

class TestWarningFilterWithWhitelist(unittest.TestCase):

    WARNINGS = [FunctionInfo("foo", 'filename'),
             FunctionInfo("bar", 'filename'),
             FunctionInfo("foo", 'anotherfile')]

    def test_should_filter_out_the_whitelist(self):
        warnings = whitelist_filter(self.WARNINGS, "foo")
        self.assertEqual(1, len(list(warnings)))

    def test_should_filter_function_in_the_right_file_when_specified(self):
        warnings = whitelist_filter(self.WARNINGS, 'filename:foo')
        self.assertEqual(2, len(list(warnings)))

    def test_should_work_with_class_member(self):
        warnings = whitelist_filter([FunctionInfo("class::foo", 'filename')], 'class::foo')
        self.assertEqual(0, len(list(warnings)))

    def test_should_filter_mutiple_functions_defined_on_one_line(self):
        warnings = whitelist_filter(self.WARNINGS, 'foo, bar')
        self.assertEqual(0, len(list(warnings)))

    def test_should_filter_mutiple_lines_of_whitelist(self):
        warnings = whitelist_filter(self.WARNINGS, 'foo\n bar')
        self.assertEqual(0, len(list(warnings)))

    def test_should_ignore_comments_in_whitelist(self):
        warnings = whitelist_filter(self.WARNINGS, 'foo  #,bar\ni#,bar')
        self.assertEqual(1, len(list(warnings)))



'''
This is an extension of lizard, that counts the amount of dependencies
within the code.
'''


class LizardExtension(object):  # pylint: disable=R0903
    FUNCTION_CAPTION = " dep cnt "
    FUNCTION_INFO_PART = "dependency_count"

    def __call__(self, tokens, reader):
        ignored_list = {','}
        dependency_type = {
            'null': 0,
            '#include': 1,
            'import': 2,
            'python_import_as_change': 3}
        expect_dependency = 0
        import_list = []
        import_as_list = []
        import_as_counter = 0
        for token in tokens:
            if not hasattr(reader.context.current_function,
                           "dependency_count"):
                reader.context.current_function.dependency_count = 0
            # this accounts for java, c, c++ and python's import
            if token == "import" or token == "#include":
                if import_as_list != []:
                    import_list.append(import_as_list)
                expect_dependency = dependency_type[token]
            elif expect_dependency == dependency_type['#include']:
                # gets rid of the <> or "" as well as the .h
                import_list += [token[1:len(token) - 3]]
                expect_dependency = dependency_type['null']
            elif expect_dependency == dependency_type['import']:
                if token == "as":
                    expect_dependency = dependency_type[
                        'python_import_as_change']
                    import_as_counter = len(import_as_list)
                    import_as_list = []
                elif import_as_counter > 4:
                    import_list += [import_as_list[0]]
                    import_as_list = []
                    import_as_counter = 0
                    expect_dependency = dependency_type['null']
                elif token not in ignored_list:
                    import_as_counter += 1
                    import_as_list += [token]
            elif (expect_dependency ==
                  dependency_type['python_import_as_change'] and
                  token not in ignored_list):
                import_as_counter -= 1
                import_list += [token]
                if import_as_counter == 0:
                    expect_dependency = dependency_type['null']
            if token in import_list:
                reader.context.current_function.dependency_count += 1
            yield token



import unittest
from test.mock import patch
from lizard_ext.lizardwordcount import LizardExtension


class FakeReader(object):

    class FI(object) : pass

    def __init__(self):
        self.fileinfo = self.FI()
        self.context = self
    def get_word_map(self):
        return self.fileinfo.wordCount


class TestWordCountPlugin(unittest.TestCase):

    def setUp(self):
        self.reader = FakeReader()
        self.ext = LizardExtension()

    def test_count_one_word(self):
        list(self.ext(["a", "b"], self.reader))
        self.assertEqual(1, self.reader.get_word_map()['a'])
        self.assertEqual(1, self.reader.get_word_map()['b'])

    def test_count_one_word_multiple_times(self):
        list(self.ext(["a", "a"], self.reader))
        self.assertEqual(2, self.reader.get_word_map()['a'])

    def test_count_one_word_multiple_times(self):
        list(self.ext(["a", "a"], self.reader))
        self.assertEqual(2, self.reader.get_word_map()['a'])

    def test_should_not_count_keywords(self):
        list(self.ext(["for"], self.reader))
        self.assertNotIn('for', self.reader.get_word_map())

    def test_should_count_non_keyword(self):
        list(self.ext(["For"], self.reader))
        self.assertIn('For', self.reader.get_word_map())

    def test_should_not_count_string(self):
        list(self.ext(["\"\""], self.reader))
        self.assertEqual(0, len(self.reader.get_word_map()))

    def test_should_not_line_continuer(self):
        list(self.ext(["\\\n"], self.reader))
        self.assertEqual(0, len(self.reader.get_word_map()))

    def test_reduce_the_result(self):
        list(self.ext(["a"], self.reader))
        self.ext.reduce(self.reader.fileinfo)
        self.ext.reduce(self.reader.fileinfo)
        self.assertEqual(2, self.ext.result['a'])

class TestWordCountOutput(unittest.TestCase):

    def setUp(self):
        self.buf = ''

    def write_to_buffer(self, txt):
            self.buf += txt

    @patch('webbrowser.open')
    @patch('lizard_ext.lizardwordcount.open', create=True)
    def test_should_output_html(self, mock_open, browser_open):
        buf = ""
        mock_open.return_value.__enter__.return_value.write.side_effect = self.write_to_buffer
        ext = LizardExtension()
        ext.result = {'a':123}
        ext.print_result()
        mock_open.assert_called_once_with('codecloud.html', 'w')
        self.assertIn('<html>', self.buf)
        self.assertIn('["a", 123]', self.buf)

    @patch('webbrowser.open')
    @patch('lizard_ext.lizardwordcount.open', create=True)
    def test_should_open_the_browser(self, mock_open, browser_open):
        import os
        ext = LizardExtension()
        ext.result = {'a':123}
        ext.print_result()
        browser_open.assert_called_with('file://' + os.path.abspath('codecloud.html'));




'''
Language parser for Apple Swift
'''

from lizard import CodeReader, CCppCommentsMixin


class SwiftReader(CodeReader, CCppCommentsMixin):
    # pylint: disable=R0903

    ext = ['swift']
    language_names = ['swift']

    def __init__(self, context):
        super(SwiftReader, self).__init__(context)

    def _state_global(self, token):
        if token == 'func':
            self._state = self._function_name
        if token == 'init':
            self._function_name(token)
        if token in ('get', 'set', 'deinit'):
            self.context.start_new_function(token)
            self._state = self._expect_function_impl
        if token == 'protocol':
            self._state = self._protocol

    def _function_name(self, token):
        self.context.start_new_function(token)
        self._state = self._expect_function_dec

    def _expect_function_dec(self, token):
        if token == '(':
            self._state = self._function_dec

    def _function_dec(self, token):
        if token == ')':
            self._state = self._expect_function_impl
        else:
            self.context.parameter(token)

    def _expect_function_impl(self, token):
        if token == '{':
            self._state = self._function_impl
            self._state(token)

    @CodeReader.read_inside_brackets_then("{}")
    def _function_impl(self, _):
        self._state = self._state_global
        self.context.end_of_function()

    @CodeReader.read_inside_brackets_then("{}")
    def _protocol(self, end_token):
        if end_token == "}":
            self._state = self._state_global



import unittest
import inspect
from lizard import  analyze_file, FileAnalyzer, get_extensions


def get_python_function_list(source_code):
    return analyze_file.analyze_source_code("a.py", source_code).function_list


class Test_parser_for_Python(unittest.TestCase):

    def test_empty_source_should_return_no_function(self):
        functions = get_python_function_list("")
        self.assertEqual(0, len(functions))

    def test_simple_python_function(self):
        class namespace1:
            def simple_function():
                if IamOnEarth:
                    return toMars()
        functions = get_python_function_list(inspect.getsource(namespace1))
        self.assertEqual(1, len(functions))
        self.assertEqual("simple_function", functions[0].name)
        self.assertEqual(2, functions[0].cyclomatic_complexity)
        self.assertEqual(4, functions[0].end_line)

    def test_parameter_count(self):
        class namespace2:
            def function_with_2_parameters(a, b):
                pass
        functions = get_python_function_list(inspect.getsource(namespace2))
        self.assertEqual(2, functions[0].parameter_count)

    def test_function_end(self):
        class namespace3:
            def simple_function(self):
                pass

            blah = 42
        functions = get_python_function_list(inspect.getsource(namespace3))
        self.assertEqual(1, len(functions))
        self.assertEqual("simple_function", functions[0].name)
        self.assertEqual(3, functions[0].end_line)

    def test_top_level_functions(self):
        functions = get_python_function_list(inspect.getsource(top_level_function_for_test))
        self.assertEqual(1, len(functions))

    def test_2_top_level_functions(self):
        functions = get_python_function_list('''
def a():
    pass
def b():
    pass
''')
        self.assertEqual(2, len(functions))
        self.assertEqual("a", functions[0].name)

    def test_2_functions(self):
        class namespace4:
            def function1(a, b):
                pass
            def function2(a, b):
                pass
        functions = get_python_function_list(inspect.getsource(namespace4))
        self.assertEqual(2, len(functions))

    def test_nested_functions(self):
        class namespace5:
            def function1(a, b):
                def function2(a, b):
                    pass
                a = 1 if b == 2 else 3
        functions = get_python_function_list(inspect.getsource(namespace5))
        self.assertEqual(2, len(functions))
        self.assertEqual("function2", functions[0].name)
        self.assertEqual(4, functions[0].end_line)
        self.assertEqual("function1", functions[1].name)
        self.assertEqual(5, functions[1].end_line)
        self.assertEqual(2, functions[1].cyclomatic_complexity)

    def test_nested_functions_ended_at_eof(self):
        class namespace6:
            def function1(a, b):
                def function2(a, b):
                    pass
        functions = get_python_function_list(inspect.getsource(namespace6))
        self.assertEqual(2, len(functions))
        self.assertEqual("function2", functions[0].name)
        self.assertEqual(4, functions[0].end_line)
        self.assertEqual("function1", functions[1].name)
        self.assertEqual(4, functions[1].end_line)

    def test_nested_functions_ended_at_same_line(self):
        class namespace7:
            def function1(a, b):
                def function2(a, b):
                    pass
            def function3():
                pass
        functions = get_python_function_list(inspect.getsource(namespace7))
        self.assertEqual(3, len(functions))
        self.assertEqual("function2", functions[0].name)
        self.assertEqual(4, functions[0].end_line)
        self.assertEqual("function1", functions[1].name)
        self.assertEqual(4, functions[1].end_line)

    def test_one_line_functions(self):
        class namespace8:
            def a( ):pass
            def b( ):pass
        functions = get_python_function_list(inspect.getsource(namespace8))
        self.assertEqual("a", functions[0].name)
        self.assertEqual("b", functions[1].name)

    def test_comment_is_not_counted_in_nloc(self):
        def function_with_comments():

            # comment
            pass
        functions = get_python_function_list(inspect.getsource(function_with_comments))
        self.assertEqual(2, functions[0].nloc)

    def test_odd_blank_line(self):
        code =  "class c:\n" + \
                "    def f():\n" +\
                "  \n" +\
                "         pass\n"
        functions = get_python_function_list(code)
        self.assertEqual(4, functions[0].end_line)

    def test_odd_line_with_comment(self):
        code =  "class c:\n" + \
                "    def f():\n" +\
                "  #\n" +\
                "         pass\n"
        functions = get_python_function_list(code)
        self.assertEqual(4, functions[0].end_line)

    def test_tab_is_same_as_8_spaces(self):
        code =  ' ' * 7 + "def a():\n" + \
                '\t'    +  "pass\n"
        functions = get_python_function_list(code)
        self.assertEqual(2, functions[0].end_line)

    def test_if_elif_and_or_for_while_except_finally(self):
        code =  'def a():\n' + \
                '    if elif and or for while except finally\n'
        functions = get_python_function_list(code)
        self.assertEqual(9, functions[0].cyclomatic_complexity)

    def test_block_string_is_one_token(self):
        code =  'def a():\n' + \
                "    a = '''\n" +\
                "a b c d e f g h i'''\n"+\
                "    return a\n"
        functions = get_python_function_list(code)
        self.assertEqual(9, functions[0].token_count)
        self.assertEqual(4, functions[0].end_line)

    def check_function_info(self, source, expect_token_count, expect_nloc, expect_endline):
        functions = get_python_function_list(source)
        self.assertEqual(expect_token_count, functions[0].token_count)
        self.assertEqual(expect_nloc, functions[0].nloc)
        self.assertEqual(expect_endline, functions[0].end_line)

    def test_block_string(self):
        self.check_function_info('def f(): a="""block string"""', 7, 1, 1)
        self.check_function_info("def f(): a='''block string'''", 7, 1, 1)
        self.check_function_info("def f():\n a='''block string'''", 7, 2, 2)
        self.check_function_info("def f():\n a='''block\n string'''", 7, 3, 3)
        self.check_function_info("def f():\n a='''block\n '''", 7, 3, 3)

    def test_docstring_is_not_counted_in_nloc(self):
        self.check_function_info("def f():\n '''block\n '''\n pass", 6, 2, 4)

    #global complexity


def top_level_function_for_test():
    pass

